@@ -1,14 +1,14 @@
 #  You may distribute under the terms of either the GNU General Public License
 #  or the Artistic License (the same terms as Perl itself)
 #
-#  (C) Paul Evans, 2011-2013 -- leonerd@leonerd.org.uk
+#  (C) Paul Evans, 2011-2014 -- leonerd@leonerd.org.uk
 
 package IO::Async::Function;
 
 use strict;
 use warnings;
 
-our $VERSION = '0.60';
+our $VERSION = '0.62';
 
 use base qw( IO::Async::Notifier );
 use IO::Async::Timer::Countdown;
@@ -39,16 +39,12 @@ C<IO::Async::Function> - call a function asynchronously
 
  $function->call(
     args => [ 123454321 ],
-    on_return => sub {
-       my $isprime = shift;
-       print "123454321 " . ( $isprime ? "is" : "is not" ) . " a prime number\n";
-    },
-    on_error => sub {
-       print STDERR "Cannot determine if it's prime - $_[0]\n";
-    },
- );
-
- $loop->run;
+ )->then( sub {
+    my $isprime = shift;
+    print "123454321 " . ( $isprime ? "is" : "is not" ) . " a prime number\n";
+ })->else(
+    print STDERR "Cannot determine if it's prime - $_[0]\n";
+ })->get;
 
 =head1 DESCRIPTION
 
@@ -113,6 +109,11 @@ The following named parameters may be passed to C<new> or C<configure>:
 
 The body of the function to execute.
 
+=item model => "spawn" | "thread"
+
+Optional. Requests a specific C<IO::Async::Routine> model. If not supplied,
+leaves the default choice up to Routine.
+
 =item min_workers => INT
 
 =item max_workers => INT
@@ -156,7 +157,7 @@ sub _init
    $self->{min_workers} = 1;
    $self->{max_workers} = 8;
 
-   $self->{workers} = {};
+   $self->{workers} = {}; # {$id} => IaFunction:Worker
 
    $self->{pending_queue} = [];
 }
@@ -167,7 +168,7 @@ sub configure
    my %params = @_;
 
    my %worker_params;
-   foreach (qw( exit_on_die max_worker_calls )) {
+   foreach (qw( model exit_on_die max_worker_calls )) {
       $self->{$_} = $worker_params{$_} = delete $params{$_} if exists $params{$_};
    }
 
@@ -193,13 +194,13 @@ sub configure
                my $workers = $self->{workers};
 
                # Shut down atmost one idle worker, starting from the highest
-               # PID. Since we search from lowest to assign work, this tries
+               # ID. Since we search from lowest to assign work, this tries
                # to ensure we'll shut down the least useful ones first,
                # keeping more useful ones in memory (page/cache warmth, etc..)
-               foreach my $pid ( reverse sort keys %$workers ) {
-                  next if $workers->{$pid}{busy};
+               foreach my $id ( reverse sort keys %$workers ) {
+                  next if $workers->{$id}{busy};
 
-                  $workers->{$pid}->stop;
+                  $workers->{$id}->stop;
                   last;
                }
 
@@ -274,6 +275,7 @@ sub stop
 {
    my $self = shift;
 
+   $self->{stopping} = 1;
    foreach my $worker ( $self->_worker_objects ) {
       $worker->stop;
    }
@@ -293,7 +295,7 @@ sub restart
    $self->start;
 }
 
-=head2 $function->call( %params )
+=head2 $function->call( %params ) ==> @result
 
 Schedules an invocation of the contained function to be executed on one of the
 worker processes. If a non-busy worker is available now, it will be called
@@ -312,6 +314,15 @@ The C<%params> hash takes the following keys:
 
 A reference to the array of arguments to pass to the code.
 
+=back
+
+=head2 $function->call( %params )
+
+When not returning a future, the C<on_result>, C<on_return> and C<on_error>
+arguments give continuations to handle successful results or failure.
+
+=over 8
+
 =item on_result => CODE
 
 A continuation that is invoked when the code has been executed. If the code
@@ -332,11 +343,6 @@ circumstances given above. They will be called directly, without the leading
 
 =back
 
-=head2 $future = $function->call( %params )
-
-When returning a future, the C<on_result>, C<on_return> and C<on_error>
-continuations are optional.
-
 =cut
 
 sub call
@@ -350,23 +356,22 @@ sub call
    my $args = delete $params{args};
    ref $args eq "ARRAY" or croak "Expected 'args' to be an array";
 
-   my $future = $self->loop->new_future;
-
+   my ( $on_done, $on_fail );
    if( defined $params{on_result} ) {
       my $on_result = delete $params{on_result};
       ref $on_result or croak "Expected 'on_result' to be a reference";
 
-      $future->on_done( $self->_capture_weakself( sub {
+      $on_done = $self->_capture_weakself( sub {
          my $self = shift or return;
-         $self->debug_printf( "CONT on_return" );
+         $self->debug_printf( "CONT on_result return" );
          $on_result->( return => @_ );
-      } ) );
-      $future->on_fail( $self->_capture_weakself( sub {
+      } );
+      $on_fail = $self->_capture_weakself( sub {
          my $self = shift or return;
          my ( $err, @values ) = @_;
-         $self->debug_printf( "CONT on_error" );
+         $self->debug_printf( "CONT on_result error" );
          $on_result->( error => @values );
-      } ) );
+      } );
    }
    elsif( defined $params{on_return} and defined $params{on_error} ) {
       my $on_return = delete $params{on_return};
@@ -374,33 +379,46 @@ sub call
       my $on_error  = delete $params{on_error};
       ref $on_error or croak "Expected 'on_error' to be a reference";
 
-      $future->on_done( $self->_capture_weakself( sub {
+      $on_done = $self->_capture_weakself( sub {
          my $self = shift or return;
          $self->debug_printf( "CONT on_return" );
          $on_return->( @_ );
-      } ) );
-      $future->on_fail( $self->_capture_weakself( sub {
+      } );
+      $on_fail = $self->_capture_weakself( sub {
          my $self = shift or return;
-         my ( $err, @values ) = @_;
          $self->debug_printf( "CONT on_error" );
-         $on_error->( @values );
-      } ) );
+         $on_error->( @_ );
+      } );
    }
    elsif( !defined wantarray ) {
       croak "Expected either 'on_result' or 'on_return' and 'on_error' keys, or to return a Future";
    }
 
-   my $worker = $self->_get_worker;
+   my $request = freeze( $args );
 
-   if( !$worker ) {
-      my $request = freeze( $args );
-      push @{ $self->{pending_queue} }, [ $request, $future ];
-      return $future;
+   my $future;
+   if( my $worker = $self->_get_worker ) {
+      $self->debug_printf( "CALL" );
+      $future = $self->_call_worker( $worker, $request );
+   }
+   else {
+      $self->debug_printf( "QUEUE" );
+      push @{ $self->{pending_queue} }, my $wait_f = $self->loop->new_future;
+
+      $future = $wait_f->then( sub {
+         my ( $self, $worker ) = @_;
+         $self->_call_worker( $worker, $request );
+      });
    }
 
-   $self->_call_worker( $worker, args => $args, $future );
+   $future->on_done( $on_done ) if $on_done;
+   $future->on_fail( $on_fail ) if $on_fail;
 
-   return $future;
+   return $future if defined wantarray;
+
+   # Caller is not going to keep hold of the Future, so we have to ensure it
+   # stays alive somehow
+   $future->on_ready( sub { undef $future } ); # intentional cycle
 }
 
 sub _worker_objects
@@ -450,13 +468,15 @@ sub _new_worker
    my $self = shift;
 
    my $worker = IO::Async::Function::Worker->new(
-      ( map { $_ => $self->{$_} } qw( code setup exit_on_die ) ),
+      ( map { $_ => $self->{$_} } qw( model code setup exit_on_die ) ),
       max_calls => $self->{max_worker_calls},
 
       on_finish => $self->_capture_weakself( sub {
          my $self = shift or return;
          my ( $worker ) = @_;
 
+         return if $self->{stopping};
+
          $self->_new_worker if $self->workers < $self->{min_workers};
 
          $self->_dispatch_pending;
@@ -465,7 +485,7 @@ sub _new_worker
 
    $self->add_child( $worker );
 
-   return $self->{workers}{$worker->pid} = $worker;
+   return $self->{workers}{$worker->id} = $worker;
 }
 
 sub _get_worker
@@ -486,24 +506,32 @@ sub _get_worker
 sub _call_worker
 {
    my $self = shift;
-   my ( $worker, $type, $args, $future ) = @_;
+   my ( $worker, $type, $args ) = @_;
 
-   $worker->call( $type, $args, $future );
+   my $future = $worker->call( $type, $args );
 
    if( $self->workers_idle == 0 ) {
       $self->{idle_timer}->stop if $self->{idle_timer};
    }
+
+   return $future;
 }
 
 sub _dispatch_pending
 {
    my $self = shift;
 
-   if( my $next = shift @{ $self->{pending_queue} } ) {
+   while( my $next = shift @{ $self->{pending_queue} } ) {
       my $worker = $self->_get_worker or return;
-      $self->_call_worker( $worker, frozen => @$next );
+
+      next if $next->is_cancelled;
+
+      $self->debug_printf( "UNQUEUE" );
+      $next->done( $self, $worker );
+      return;
    }
-   elsif( $self->workers_idle > $self->{min_workers} ) {
+
+   if( $self->workers_idle > $self->{min_workers} ) {
       $self->{idle_timer}->start if $self->{idle_timer} and !$self->{idle_timer}->is_running;
    }
 }
@@ -533,7 +561,8 @@ sub new
             $ret_channel->send( [ r => @ret ] );
          }
          else {
-            $ret_channel->send( [ e => "$@" ] );
+            chomp( my $e = "$@" );
+            $ret_channel->send( [ e => $e, error => ] );
          }
       }
    };
@@ -566,65 +595,54 @@ sub stop
    $worker->{arg_channel}->close;
 
    if( my $function = $worker->parent ) {
-      delete $function->{workers}{$worker->pid};
+      delete $function->{workers}{$worker->id};
    }
 }
 
 sub call
 {
    my $worker = shift;
-   my ( $type, $args, $future ) = @_;
+   my ( $args ) = @_;
 
-   if( $type eq "args" ) {
-      $worker->{arg_channel}->send( $args );
-   }
-   elsif( $type eq "frozen" ) {
-      $worker->{arg_channel}->send_frozen( $args );
-   }
-   else {
-      die "TODO: unsure $type\n";
-   }
+   $worker->{arg_channel}->send_frozen( $args );
 
-   $worker->{ret_channel}->recv(
-      on_recv => $worker->_capture_weakself( sub {
-         my ( $worker, $channel, $result ) = @_;
+   $worker->{busy} = 1;
+   $worker->{max_calls}--;
+
+   return $worker->{ret_channel}->recv->then(
+      # on recv
+      $worker->_capture_weakself( sub {
+         my ( $worker, $result ) = @_;
          my ( $type, @values ) = @$result;
 
-         $worker->{busy} = 0;
-
-         my $function = $worker->parent;
+         $worker->stop if !$worker->{max_calls} or
+                          $worker->{exit_on_die} && $type eq "e";
 
          if( $type eq "r" ) {
-            $future->done( @values );
+            return Future->new->done( @values );
          }
          elsif( $type eq "e" ) {
-            $future->fail( $values[0], @values );
-            $worker->stop if $worker->{exit_on_die};
+            return Future->new->fail( @values );
          }
          else {
             die "Unrecognised type from worker - $type\n";
          }
-
-         $worker->stop if !$worker->{max_calls};
-
-         $function->_dispatch_pending if $function;
       } ),
-      on_eof => $worker->_capture_weakself( sub {
-         my ( $worker, $channel ) = @_;
+      # on EOF
+      $worker->_capture_weakself( sub {
+         my ( $worker ) = @_;
 
-         $worker->{busy} = 0;
-
-         my $function = $worker->parent;
-
-         $future->fail( "closed", "closed" );
          $worker->stop;
 
-         $function->_dispatch_pending if $function;
-      } ),
-   );
+         return Future->new->fail( "closed", "closed" );
+      } )
+   )->on_ready( $worker->_capture_weakself( sub {
+      my ( $worker, $f ) = @_;
+      $worker->{busy} = 0;
 
-   $worker->{busy} = 1;
-   $worker->{max_calls}--;
+      my $function = $worker->parent;
+      $function->_dispatch_pending if $function;
+   }));
 }
 
 =head1 NOTES
