@@ -8,7 +8,7 @@ package IO::Async::Notifier;
 use strict;
 use warnings;
 
-our $VERSION = '0.60';
+our $VERSION = '0.62';
 
 use Carp;
 use Scalar::Util qw( weaken );
@@ -17,6 +17,9 @@ use Scalar::Util qw( weaken );
 use constant HAS_BROKEN_TRAMPOLINES => ( $] == "5.008004" );
 
 our $DEBUG = $ENV{IO_ASYNC_DEBUG} || 0;
+our $DEBUG_FD   = $ENV{IO_ASYNC_DEBUG_FD};
+our $DEBUG_FILE = $ENV{IO_ASYNC_DEBUG_FILE};
+our $DEBUG_FH;
 
 =head1 NAME
 
@@ -655,7 +658,7 @@ sub make_event_cb
    return $self->_capture_weakself( 
       !$DEBUG ? $code : sub {
          my $self = $_[0];
-         $self->debug_printf_event( $caller, $event_name );
+         $self->_debug_printf_event( $caller, $event_name );
          goto &$code;
       }
    );
@@ -681,7 +684,7 @@ sub maybe_make_event_cb
    return $self->_capture_weakself(
       !$DEBUG ? $code : sub {
          my $self = $_[0];
-         $self->debug_printf_event( $caller, $event_name );
+         $self->_debug_printf_event( $caller, $event_name );
          goto &$code;
       }
    );
@@ -704,7 +707,7 @@ sub invoke_event
    my $code = $self->can_event( $event_name )
       or croak "$self cannot handle $event_name event";
 
-   $self->debug_printf_event( scalar caller, $event_name ) if $DEBUG;
+   $self->_debug_printf_event( scalar caller, $event_name ) if $DEBUG;
    return $code->( $self, @args );
 }
 
@@ -726,7 +729,7 @@ sub maybe_invoke_event
    my $code = $self->can_event( $event_name )
       or return undef;
 
-   $self->debug_printf_event( scalar caller, $event_name ) if $DEBUG;
+   $self->_debug_printf_event( scalar caller, $event_name ) if $DEBUG;
    return [ $code->( $self, @args ) ];
 }
 
@@ -743,6 +746,20 @@ When debugging is enabled, the C<make_event_cb> and C<invoke_event> methods
 a debugging line is printed, using the C<debug_printf> method. This identifes
 the name of the event.
 
+By default, the line is only printed if the caller of one of these methods is
+the same package as the object is blessed into, allowing it to print the
+events of the most-derived class, without the extra verbosity of the
+lower-level events of its parent class used to create it. All calls regardless
+of caller can be printed by setting a number greater than 1 as the value of
+C<IO_ASYNC_DEBUG>.
+
+By default the debugging log goes to C<STDERR>, but two other environment
+variables can redirect it. If C<IO_ASYNC_DEBUG_FILE> is set, it names a file
+which will be opened for writing, and logging written into it. Otherwise, if
+C<IO_ASYNC_DEBUG_FD> is set, it gives a file descriptor number that logging
+should be written to. If opening the named file or file descriptor fails then
+the log will be written to C<STDERR> as normal.
+
 =cut
 
 =head2 $notifier->debug_printf( $format, @args )
@@ -795,11 +812,25 @@ sub debug_printf
    s/^IO::Async::/Ia:/,
    s/^Net::Async::/Na:/ for @id;
 
-   printf STDERR "[%s] $format\n",
+   $DEBUG_FH ||= do {
+      my $fh;
+      if( $DEBUG_FILE ) {
+         open $fh, ">", $DEBUG_FILE or undef $fh;
+      }
+      elsif( $DEBUG_FD ) {
+         $fh = IO::Handle->new;
+         $fh->fdopen( $DEBUG_FD, "w" ) or undef $fh;
+      }
+      $fh ||= \*STDERR;
+      $fh->autoflush;
+      $fh;
+   };
+
+   printf $DEBUG_FH "[%s] $format\n",
       join("<-", @id), @args;
 }
 
-sub debug_printf_event
+sub _debug_printf_event
 {
    my $self = shift;
    my ( $caller, $event_name ) = @_;
