@@ -8,7 +8,7 @@ package IO::Async::OS::MSWin32;
 use strict;
 use warnings;
 
-our $VERSION = '0.60';
+our $VERSION = '0.62';
 
 our @ISA = qw( IO::Async::OS::_Base );
 
@@ -31,6 +31,12 @@ use constant HAVE_RENAME_OPEN_FILES => 0;
 # well try the Select loop first
 use constant LOOP_BUILTIN_CLASSES => qw( Select Poll );
 
+# CORE::fork() does not provide full POSIX semantics
+use constant HAVE_POSIX_FORK => 0;
+
+# Windows does not have signals, and SIGCHLD is not available
+use constant HAVE_SIGNALS => 0;
+
 =head1 NAME
 
 C<IO::Async::OS::MSWin32> - operating system abstractions on C<MSWin32> for C<IO::Async>
