@@ -9,24 +9,23 @@ use strict;
 use warnings;
 use base qw( IO::Async::Function );
 
-our $VERSION = '0.60';
+our $VERSION = '0.62';
 
-use Socket 1.93 qw(
+# Socket 2.006 fails to getaddrinfo() AI_NUMERICHOST properly on MSWin32
+use Socket 2.007 qw(
    AI_NUMERICHOST AI_PASSIVE
    NI_NUMERICHOST NI_NUMERICSERV NI_DGRAM
    EAI_NONAME
 );
 
-BEGIN {
-   # We're going to implement methods called getaddrinfo and getnameinfo. We
-   # therefore need import them with different names prefixed with underscores
-   *_getaddrinfo = \&Socket::getaddrinfo;
-   *_getnameinfo = \&Socket::getnameinfo;
-}
-
 use IO::Async::OS;
 
-use Time::HiRes qw( alarm );
+# Try to use HiRes alarm, but we don't strictly need it.
+# MSWin32 doesn't implement it
+BEGIN {
+   require Time::HiRes;
+   eval { Time::HiRes::alarm(0) } and Time::HiRes->import( qw( alarm ) );
+}
 
 use Carp;
 
@@ -47,29 +46,18 @@ This object is used indirectly via an C<IO::Async::Loop>:
  $loop->resolver->getaddrinfo(
     host    => "www.example.com",
     service => "http",
+ )->on_done( sub {
+    foreach my $addr ( @_ ) {
+       printf "http://www.example.com can be reached at " .
+          "socket(%d,%d,%d) + connect('%v02x')\n",
+          @{$addr}{qw( family socktype protocol addr )};
+    }
+ });
 
-    on_resolved => sub {
-       foreach my $addr ( @_ ) {
-          printf "http://www.example.com can be reached at " .
-             "socket(%d,%d,%d) + connect('%v02x')\n",
-             @{$addr}{qw( family socktype protocol addr )};
-       }
-    },
-
-    on_error => sub {
-       print "Cannot look up www.example.com - $_[-1]\n";
-    },
- );
-
- $loop->resolve( type => 'getpwuid', data => [ $< ],
-    on_resolved => sub {
-       print "My passwd ent: " . join( "|", @_ ) . "\n";
-    },
-
-    on_error => sub {
-       print "Cannot look up my passwd ent - $_[-1]\n";
-    },
- );
+ $loop->resolve( type => 'getpwuid', data => [ $< ] )
+    ->on_done( sub {
+    print "My passwd ent: " . join( "|", @_ ) . "\n";
+ });
 
  $loop->run;
 
@@ -129,7 +117,7 @@ sub _init
 
 =cut
 
-=head2 $loop->resolve( %params )
+=head2 $loop->resolve( %params ) ==> @result
 
 Performs a single name resolution operation, as given by the keys in the hash.
 
@@ -152,11 +140,22 @@ specific function chosen by the C<type>; see BUILT-IN RESOLVERS.
 Optional. Timeout in seconds, after which the resolver operation will abort
 with a timeout exception. If not supplied, a default of 10 seconds will apply.
 
+=back
+
+=head2 $resolver->resolve( %params )
+
+When not returning a future, additional parameters can be given containing the
+continuations to invoke on success or failure:
+
+=over 8
+
 =item on_resolved => CODE
 
 A continuation that is invoked when the resolver function returns a successful
 result. It will be passed the array returned by the resolver function.
 
+ $on_resolved->( @result )
+
 =item on_error => CODE
 
 A continuation that is invoked when the resolver function fails. It will be
@@ -164,11 +163,6 @@ passed the exception thrown by the function.
 
 =back
 
-=head2 $future = $resolver->resolve( %params )
-
-When returning a future, the C<on_resolved> and C<on_error> continuations are
-optional.
-
 =cut
 
 sub resolve
@@ -179,10 +173,8 @@ sub resolve
    my $type = $args{type};
    defined $type or croak "Expected 'type'";
 
-   # Legacy
    if( $type eq "getaddrinfo" ) {
-      warnings::warnif( deprecated => "getaddrinfo resolver will be changed in a future release to be the same as getaddrinfo_hash; please update your code to call it specifically" );
-      $type = "getaddrinfo_array";
+      $type = "getaddrinfo_hash";
    }
 
    exists $METHODS{$type} or croak "Expected 'type' to be an existing resolver method, got '$type'";
@@ -212,10 +204,14 @@ sub resolve
    $future->on_done( $on_resolved ) if $on_resolved;
    $future->on_fail( $on_error    ) if $on_error;
 
-   return $future;
+   return $future if defined wantarray;
+
+   # Caller is not going to keep hold of the Future, so we have to ensure it
+   # stays alive somehow
+   $future->on_ready( sub { undef $future } ); # intentional cycle
 }
 
-=head2 $resolver->getaddrinfo( %args )
+=head2 $resolver->getaddrinfo( %args ) ==> @addrs
 
 A shortcut wrapper around the C<getaddrinfo> resolver, taking its arguments in
 a more convenient form.
@@ -251,12 +247,32 @@ C<Socket>.
 
 Time in seconds after which to abort the lookup with a C<Timed out> exception
 
+=back
+
+On success, the future will yield the result as a list of HASH references;
+each containing one result. Each result will contain fields called C<family>,
+C<socktype>, C<protocol> and C<addr>. If requested by C<AI_CANONNAME> then the
+C<canonname> field will also be present.
+
+As a specific optimisation, this method will try to perform a lookup of
+numeric values synchronously, rather than asynchronously, if it looks likely
+to succeed.
+
+Specifically, if the service name is entirely numeric, and the hostname looks
+like an IPv4 or IPv6 string, a synchronous lookup will first be performed
+using the C<AI_NUMERICHOST> flag. If this gives an C<EAI_NONAME> error, then
+the lookup is performed asynchronously instead.
+
+=head2 $resolver->getaddrinfo( %args )
+
+When not returning a future, additional parameters can be given containing the
+continuations to invoke on success or failure:
+
+=over 8
+
 =item on_resolved => CODE
 
-Callback which is invoked after a successful lookup. Will be passed a list of
-HASH references; each containing one result. Each result will contain fields
-called C<family>, C<socktype>, C<protocol> and C<addr>. If requested by
-C<AI_CANONNAME> then the C<canonname> field will also be valid.
+Callback which is invoked after a successful lookup.
 
  $on_resolved->( @addrs )
 
@@ -268,20 +284,6 @@ Callback which is invoked after a failed lookup, including for a timeout.
 
 =back
 
-As a specific optimsation, this method will try to perform a lookup of numeric
-values synchronously, rather than asynchronously, if it looks likely to
-succeed.
-
-Specifically, if the service name is entirely numeric, and the hostname looks
-like an IPv4 or IPv6 string, a synchronous lookup will first be performed
-using the C<AI_NUMERICHOST> flag. If this gives an C<EAI_NONAME> error, then
-the lookup is performed asynchronously instead.
-
-=head2 $future = $resolver->getaddrinfo( %args )
-
-When returning a future, the C<on_resolved> and C<on_error> continuations are
-optional.
-
 =cut
 
 sub getaddrinfo
@@ -317,7 +319,7 @@ sub getaddrinfo
    if( ( $host =~ m/^[\d.]+$/ or $host =~ m/^[[:xdigit:]:]$/ or $host eq "" ) and
        $service =~ m/^\d+$/ ) {
 
-       my ( $err, @results ) = _getaddrinfo( $host, $service,
+       my ( $err, @results ) = Socket::getaddrinfo( $host, $service,
           { %args, flags => $flags | AI_NUMERICHOST }
        );
 
@@ -345,15 +347,22 @@ sub getaddrinfo
          map { exists $args{$_} ? ( $_ => $args{$_} ) : () } qw( family socktype protocol ),
       ],
       timeout => $args{timeout},
-   );
+   )->else( sub {
+      my $message = shift;
+      Future->new->fail( $message, resolve => getaddrinfo => @_ );
+   });
 
    $future->on_done( $args{on_resolved} ) if $args{on_resolved};
    $future->on_fail( $args{on_error}    ) if $args{on_error};
 
-   return $future;
+   return $future if defined wantarray;
+
+   # Caller is not going to keep hold of the Future, so we have to ensure it
+   # stays alive somehow
+   $future->on_ready( sub { undef $future } ); # intentional cycle
 }
 
-=head2 $resolver->getnameinfo( %args )
+=head2 $resolver->getnameinfo( %args ) ==> ( $host, $service )
 
 A shortcut wrapper around the C<getnameinfo> resolver, taking its arguments in
 a more convenient form.
@@ -385,9 +394,22 @@ If true, sets both C<NI_NUMERICHOST> and C<NI_NUMERICSERV> flags.
 
 Time in seconds after which to abort the lookup with a C<Timed out> exception
 
+=back
+
+As a specific optimisation, this method will try to perform a lookup of
+numeric values synchronously, rather than asynchronously, if both the
+C<NI_NUMERICHOST> and C<NI_NUMERICSERV> flags are given.
+
+=head2 $resolver->getnameinfo( %args )
+
+When not returning a future, additional parameters can be given containing the
+continuations to invoke on success or failure:
+
+=over 8
+
 =item on_resolved => CODE
 
-Callback which is invoked after a successful lookup. 
+Callback which is invoked after a successful lookup.
 
  $on_resolved->( $host, $service )
 
@@ -399,15 +421,6 @@ Callback which is invoked after a failed lookup, including for a timeout.
 
 =back
 
-As a specific optimsation, this method will try to perform a lookup of numeric
-values synchronously, rather than asynchronously, if both the
-C<NI_NUMERICHOST> and C<NI_NUMERICSERV> flags are given.
-
-=head2 $future = $resolver->getnameinfo( %args )
-
-When returning a future, the C<on_resolved> and C<on_error> continuations are
-optional.
-
 =cut
 
 sub getnameinfo
@@ -431,7 +444,7 @@ sub getnameinfo
 
    if( $flags & (NI_NUMERICHOST|NI_NUMERICSERV) ) {
       # This is a numeric-only lookup that can be done synchronously
-      my ( $err, $host, $service ) = _getnameinfo( $args{addr}, $flags );
+      my ( $err, $host, $service ) = Socket::getnameinfo( $args{addr}, $flags );
 
       if( $err ) {
          my $future = $self->loop->new_future->fail( $err, resolve => getnameinfo => $err+0 );
@@ -451,12 +464,19 @@ sub getnameinfo
       timeout => $args{timeout},
    )->transform(
       done => sub { @{ $_[0] } }, # unpack the ARRAY ref
-   );
+   )->else( sub {
+      my $message = shift;
+      Future->new->fail( $message, resolve => getnameinfo => @_ );
+   });
 
    $future->on_done( $args{on_resolved} ) if $args{on_resolved};
    $future->on_fail( $args{on_error}    ) if $args{on_error};
 
-   return $future;
+   return $future if defined wantarray;
+
+   # Caller is not going to keep hold of the Future, so we have to ensure it
+   # stays alive somehow
+   $future->on_ready( sub { undef $future } ); # intentional cycle
 }
 
 =head1 FUNCTIONS
@@ -546,7 +566,8 @@ The C<getaddrinfo_hash> resolver takes arguments in a hash of name/value pairs
 and returns a list of hash structures, as the C<Socket::getaddrinfo> function
 does. For neatness it takes all its arguments as named values; taking the host
 and service names from arguments called C<host> and C<service> respectively;
-all the remaining arguments are passed into the hints hash.
+all the remaining arguments are passed into the hints hash. This name is also
+aliased as simply C<getaddrinfo>.
 
 The C<getaddrinfo_array> resolver behaves more like the C<Socket6> version of
 the function. It takes hints in a flat list, and mangles the result of the
@@ -559,12 +580,6 @@ string names for the C<family> argument, converting C<inet> and possibly
 C<inet6> into the appropriate C<AF_*> value, and for the C<socktype> argument,
 converting C<stream>, C<dgram> or C<raw> into the appropriate C<SOCK_*> value.
 
-For backward-compatibility with older code, the resolver name C<getaddrinfo>
-is currently aliased to C<getaddrinfo_array>; but any code that wishes to rely
-on the array-like nature of its arguments and return values, should request it
-specifically by name, as this alias will be changed in a later version of
-C<IO::Async>.
-
 The C<getnameinfo> resolver returns its result in the same form as C<Socket>.
 
 Because this module simply uses the system's C<getaddrinfo> resolver, it will
@@ -585,7 +600,7 @@ register_resolver getaddrinfo_hash => sub {
    # Clear any other existing but undefined hints
    defined $args{$_} or delete $args{$_} for keys %args;
 
-   my ( $err, @addrs ) = _getaddrinfo( $host, $service, \%args );
+   my ( $err, @addrs ) = Socket::getaddrinfo( $host, $service, \%args );
 
    die "$err\n" if $err;
 
@@ -604,7 +619,7 @@ register_resolver getaddrinfo_array => sub {
    $hints{protocol} = $protocol if defined $protocol;
    $hints{flags}    = $flags    if defined $flags;
 
-   my ( $err, @addrs ) = _getaddrinfo( $host, $service, \%hints );
+   my ( $err, @addrs ) = Socket::getaddrinfo( $host, $service, \%hints );
 
    die "$err\n" if $err;
 
@@ -617,7 +632,7 @@ register_resolver getaddrinfo_array => sub {
 register_resolver getnameinfo => sub {
    my ( $addr, $flags ) = @_;
 
-   my ( $err, $host, $service ) = _getnameinfo( $addr, $flags || 0 );
+   my ( $err, $host, $service ) = Socket::getnameinfo( $addr, $flags || 0 );
 
    die "$err\n" if $err;
 
